##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Basic schema implementations.

$Id: schemas.py,v 1.1 2004/06/28 18:08:29 johanc Exp $
"""

from types import StringType

ok_types = ['unicode', 'string', 'int', 'float', 'bool', 'object',
            'classification', 'string:list', 'blob']


def add_field_type(t):
    """Adds an allowable field type."""
    assert isinstance(t, StringType)
    if t not in ok_types:
        ok_types.append(t)


class FieldSchema:
    """Defines the schema of one field."""
#    __implements__ = IRelationalSchema

    def __init__(self, name, type='string', unique=0):
        assert type in ok_types, type
        self.name = name
        self.type = type
        self.unique = not not unique

    def get_column_defs(self):
        return [(self.name, self.type, self.unique)]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if (other.name == self.name) and (other.type == self.type) and (
                other.unique == self.unique):
                return 1  # Same
        return 0  # Different

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'FieldSchema(%s, %s, %s)' % (
            repr(self.name), repr(self.type), repr(self.unique))


class RowSchema:
    """Defines an ordered set of fields for exactly one row.
    """
#    __implements__ = IRelationalSchema

    def __init__(self, fields=()):
        self.fields = []
        self.field_names = {}
        for c in fields:
            self._add(c)

    def get_column_defs(self):
        res = []
        for f in self.fields:
            res.extend(f.get_column_defs())
        return tuple(res)

    def _add(self, c):
        if self.field_names.has_key(c.name):
            raise KeyError, 'Duplicate field name: %s' % c.name
        self.field_names[c.name] = 1
        self.fields.append(c)

    def add(self, name, type='string', unique=0):
        self._add(FieldSchema(name, type, unique))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if (self.fields == other.fields):
                return 1  # Same
        return 0  # Different

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'RowSchema(%s)' % repr(self.fields)


class RowSequenceSchema (RowSchema):
    """Defines a schema for a sequence of rows, including row count limits.
    """

    def __init__(self, fields=(), min_rows=0, max_rows=0):
        # max_rows == 0 means unlimited.
        assert (max_rows == 0 or max_rows >= min_rows)
        RowSchema.__init__(self, fields)
        self.min_rows = min_rows
        self.max_rows = max_rows

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if (self.fields == other.fields) and (
                self.min_rows == other.min_rows) and (
                self.max_rows == other.max_rows):
                return 1  # Same
        return 0  # Different

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'RowSequenceSchema(%s, min_rows=%s, max_rows=%s)' % (
            repr(self.fields), repr(self.min_rows), repr(self.max_rows))

