##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""SQL database connections via DB-API 2.0.

$Id: dbapi.py,v 1.1 2004/06/28 18:08:29 johanc Exp $
"""

import os
import re
from time import time

from apelib.core.interfaces import ITPCConnection

from apelib.sql.interfaces import IRDBMSConnection

name_style_re = re.compile(':[A-Za-z0-9_-]+')

DEBUG = os.environ.get('APE_DEBUG_SQL')
PROFILE = os.environ.get('APE_PROFILE_SQL')


class AbstractSQLConnection:

    __implements__ = IRDBMSConnection, ITPCConnection

    column_type_translations = {}  # { local type name -> db type name }
    column_name_translations = {}  # { local col name -> db col name }
    module = None
    connector = None

    def __init__(self, module_name, connect_expression, prefix=''):
        # connect_expression is a Python expression.
        self.module_name = module_name
        self.module = __import__(module_name, {}, {}, ('__doc__',))
        self.connect_expression = connect_expression
        self.prefix = prefix
        self.connector = None
        self._final = 0

    def __repr__(self):
        return '<%s(module_name=%s)>' % (
            self.__class__.__name__, repr(self.module_name))

    def translate_name(self, column_name):
        """Returns a column name for a variable name.

        Defaults to no translation.
        """
        return self.column_name_translations.get(column_name, column_name)

    def translate_type(self, column_type):
        """Returns a database type for a variable type.

        If the type is unknown, raises KeyError.
        """
        return self.column_type_translations.get(column_type, column_type)

    def generate_conditions(self, keys):
        conditions = []
        for key in keys:
            clause = "%s = :%s" % (self.translate_name(key), key)
            conditions.append(clause)
        return ' AND '.join(conditions)

    def generate_insert(self, table_name, columns):
        dbcols = [self.translate_name(col) for col in columns]
        colfmts = [':%s' % col for col in columns]
        return 'INSERT INTO %s (%s) VALUES (%s)' % (
            table_name, ', '.join(dbcols), ', '.join(colfmts))

    def generate_update(self, table_name, key_columns, other_columns):
        where = self.generate_conditions(key_columns)
        to_set = [
            ("%s = :%s" % (self.translate_name(col), col))
            for col in other_columns]
        return 'UPDATE %s SET %s WHERE %s' % (
            table_name, ', '.join(to_set), where)

    def generate_delete(self, table_name, keys):
        where = self.generate_conditions(keys)
        sql = 'DELETE FROM %s' % table_name
        if where:
            sql += ' WHERE %s' % where
        return sql

    def make_dict(self, columns, data, oid=None):
        res = {}
        for n in range(len(columns)):
            res[columns[n]] = data[n]
        if oid is not None:
            res['oid'] = oid
        return res

    #
    # IRDBMSConnection implementation.
    #

    def select(self, table, result_columns, **filter):
        """Selects rows from a table and returns column values for those rows.
        """
        table_name = self.prefix + table
        dbcols = [self.translate_name(col) for col in result_columns]
        where = self.generate_conditions(filter.keys())
        sql = 'SELECT %s FROM %s' % (', '.join(dbcols), table_name)
        if where:
            sql += ' WHERE %s' % where
        return self.execute(sql, filter, fetch=1)

    def insert(self, table, columns, row):
        """Inserts one row in the table.
        """
        table_name = self.prefix + table
        kw = self.make_dict(columns, row)
        sql = self.generate_insert(table_name, columns)
        self.execute(sql, kw)

    def set_one(self, table, oid, columns, row, is_new):
        """Sets one row in a table.

        Requires the table to have only one value for each oid.
        Executes either an update or insert operation, depending on
        the is_new argument and configured policies.
        """
        table_name = self.prefix + table
        kw = self.make_dict(columns, row, oid)
        if is_new:
            sql = self.generate_insert(table_name, ('oid',) + tuple(columns))
            self.execute(sql, kw)
        else:
            sql = self.generate_update(table_name, ('oid',), columns)
            self.execute(sql, kw)
            
    def set_many(self, table, oid, key_columns, other_columns, rows):
        """Sets multiple rows in a table.

        'rows' is a sequence of tuples containing values for the
        key_columns as well as the other_columns.

        Either deletes all rows for an oid and inserts new rows, or
        examines the current state of the database and modifies it in
        pieces.
        """
        table_name = self.prefix + table
        combo = tuple(key_columns) + tuple(other_columns)
        if not key_columns:
            # Don't compare rows.  Just delete and insert.
            sql = self.generate_delete(table_name, ('oid',))
            self.execute(sql, {'oid': oid})
            for row in rows:
                sql = self.generate_insert(table_name, ('oid',) + combo)
                kw = self.make_dict(combo, row, oid)
                self.execute(sql, kw)
            return
        # Edit the table.
        exist_rows = self.select(table, combo, oid=oid)
        count = len(key_columns)
        existing = {}
        for record in exist_rows:
            key = tuple(record[:count])
            value = tuple(record[count:])
            existing[key] = value
        now = {}
        for record in rows:
            key = tuple(record[:count])
            value = tuple(record[count:])
            now[key] = value
        # Delete and update rows.
        for key, value in existing.items():
            if not now.has_key(key):
                # Delete this row.
                sql = self.generate_delete(
                    table_name, ('oid',) + tuple(key_columns))
                kw = self.make_dict(key_columns, key, oid)
                self.execute(sql, kw)
            elif now[key] != value:
                # Update this row.
                #print 'DIFFERENT:', now[key], value
                sql = self.generate_update(
                    table_name, ('oid',) + tuple(key_columns), other_columns)
                kw = self.make_dict(combo, key + now[key], oid)
                self.execute(sql, kw)
        for key, value in now.items():
            if not existing.has_key(key):
                # Insert this row.
                sql = self.generate_insert(table_name, ('oid',) + combo)
                kw = self.make_dict(combo, key + value, oid)
                self.execute(sql, kw)
        return

    def delete_from(self, table, **filter):
        """Deletes rows from a table.
        """
        table_name = self.prefix + table
        sql = self.generate_delete(table_name, filter.keys())
        self.execute(sql, filter)

    def exists(self, name, type_name):
        """Returns true if the specified database object exists.

        type_name is 'table' or 'sequence'
        """
        raise NotImplementedError("Abstract Method")

    def list_table_names(self):
        """Returns a list of existing table names.
        """
        raise NotImplementedError("Abstract Method")

    def create_table(self, table, column_defs):
        """Creates a table.
        """
        table_name = self.prefix + table
        cols = []
        pkeys = []
        for name, typ, unique in column_defs:
            col = self.translate_name(name)
            db_type = self.translate_type(typ)
            cols.append("%s %s" % (col, db_type))
            if unique:
                pkeys.append(col)
        if pkeys:
            cols.append('PRIMARY KEY (%s)' % ', '.join(pkeys))
        sql = "CREATE TABLE %s (%s)" % (table_name, ', '.join(cols))
        self.execute(sql)

    def drop_table(self, table):
        """Drops a table.
        """
        table_name = self.prefix + table
        sql = "DROP TABLE %s" % table_name
        self.execute(sql)

    def create_sequence(self, name, start=1):
        """Creates a sequence.
        """
        raise NotImplementedError("Abstract Method")

    def reset_sequence(self, name, start=1):
        """Resets a sequence.
        """
        raise NotImplementedError("Abstract Method")

    def increment(self, name):
        """Increments a sequence.
        """
        raise NotImplementedError("Abstract Method")

    def execute(self, sql, kw=None, fetch=0):
        if self.connector is None:
            raise RuntimeError('Not connected')
        if kw is None:
            kw = {}
        converter = style_converters[self.module.paramstyle]
        sql, params = converter(sql, kw)
        cursor = self.connector.cursor()
        try:
            if DEBUG or PROFILE:
                print 'SQL: %s, %s' % (repr(sql), params)
            if PROFILE:
                start = time()
                cursor.execute(sql, params)
                end = time()
                print 'SQL time: %0.6fs' % (end - start)
            else:
                cursor.execute(sql, params)
            if fetch:
                res = list(cursor.fetchall())
                if DEBUG:
                    print 'SQL result: %s' % repr(res)
                return res
        finally:
            cursor.close()

    #
    # ITPCConnection implementation.
    #

    def connect(self):
        d = {'connect': self.module.connect}
        self.connector = eval(self.connect_expression, d)

    def sortKey(self):
        return repr(self)

    def getName(self):
        return repr(self)

    def begin(self):
        pass

    def vote(self):
        self._final = 1

    def reset(self):
        self._final = 0

    def abort(self):
        try:
            self.connector.rollback()
        finally:
            self.reset()

    def finish(self):
        if self._final:
            try:
                self.connector.commit()
            finally:
                self.reset()

    def close(self):
        c = self.connector
        if c is not None:
            self.connector = None
            c.close()



# Converters for all parameter styles defined by DB-API 2.0

style_converters = {}

def convert_to_qmark(sql, kw):
    # '?' format
    params = []
    def replace(match, params=params, kw=kw):
        name = match.group()[1:]
        params.append(kw[name])
        return '?'
    sql = name_style_re.sub(replace, sql)
    return sql, tuple(params)
style_converters['qmark'] = convert_to_qmark

def convert_to_numeric(sql, kw):
    # ':1' format
    params = []
    def replace(match, params=params, kw=kw):
        name = match.group()[1:]
        index = len(params)
        params.append(kw[name])
        return ':%d' % index
    sql = name_style_re.sub(replace, sql)
    return sql, tuple(params)
style_converters['numeric'] = convert_to_numeric

def convert_to_named(sql, kw):
    # ':name' format
    # The input format is the same as the output format.
    return sql, kw
style_converters['named'] = convert_to_named

def convert_to_format(sql, kw):
    # '%s' format
    params = []
    def replace(match, params=params, kw=kw):
        name = match.group()[1:]
        params.append(kw[name])
        return '%s'
    sql = name_style_re.sub(replace, sql)
    return sql, tuple(params)
style_converters['format'] = convert_to_format

def convert_to_pyformat(sql, kw):
    # '%(name)s' format
    def replace(match, kw=kw):
        name = match.group()[1:]
        return '%%(%s)s' % name
    sql = name_style_re.sub(replace, sql)
    return sql, kw
style_converters['pyformat'] = convert_to_pyformat


# Database-specific implementations of IRDBMSConnection.

class PostgreSQLConnection (AbstractSQLConnection):

    column_type_translations = {
        'long':   'bigint',
        'string': 'character varying(255)',
        'blob':   'bytea',
        'datetime': 'timestamp',
        }

    column_name_translations = {
        'oid': 'objoid',
        }

    def exists(self, name, type_name):
        """Returns true if the specified database object exists.

        type_name is 'table' or 'sequence'
        """
        table_name = self.prefix + name
        if type_name == 'table':
            sql = ('SELECT tablename FROM pg_tables '
                   'WHERE tablename = :name')
        elif type_name == 'sequence':
            sql = ("SELECT relname FROM pg_class "
                   "WHERE relkind = 'S' AND relname = :name")
        else:
            raise ValueError(type_name)
        rows = self.execute(sql, {'name': table_name.lower()}, fetch=1)
        return len(rows)

    def list_table_names(self):
        """Returns a list of existing table names.
        """
        sql = 'SELECT tablename FROM pg_tables'
        rows = self.execute(sql, {}, fetch=1)
        res = []
        for (name,) in rows:
            if not self.prefix or name.startswith(self.prefix):
                res.append(name[len(self.prefix):])
        return res

    def create_sequence(self, name, start=1):
        """Creates a sequence.
        """
        sql = "CREATE SEQUENCE %s START %d" % (self.prefix + name, start)
        self.execute(sql)

    def reset_sequence(self, name, start=1):
        """Resets a sequence.
        """
        sql = "SELECT setval('%s', %d)" % (self.prefix + name, start)
        self.execute(sql)

    def increment(self, name):
        """Increments a sequence.
        """
        sql = "SELECT nextval('%s')" % (self.prefix + name)
        rows = self.execute(sql, fetch=1)
        return rows[0][0]



class MySQLConnection (AbstractSQLConnection):

    column_type_translations = {
        'long':   'bigint',
        'string': 'character varying(255)',
        'blob':   'longblob',
        'boolean': 'tinyint(1)',
        }

    column_name_translations = {
        'oid': 'objoid',
        }

    def exists(self, name, type_name):
        """Returns true if the specified database object exists.

        type_name is 'table' or 'sequence'
        """
        table_name = self.prefix + name
        if type_name not in ('table', 'sequence'):
            raise ValueError(type_name)
        sql = 'SHOW TABLES LIKE :name'
        rows = self.execute(sql, {'name': table_name}, fetch=1)
        return len(rows)

    def list_table_names(self):
        """Returns a list of existing table names.
        """
        sql = 'SHOW TABLES'
        rows = self.execute(sql, {}, fetch=1)
        res = []
        for (name,) in rows:
            if not self.prefix or name.startswith(self.prefix):
                res.append(name[len(self.prefix):])
        return res

    def create_sequence(self, name, start=1):
        """Creates a sequence.
        """
        table_name = self.prefix + name
        self.execute("CREATE TABLE %s (last_value int)" % table_name)
        self.execute("INSERT INTO %s VALUES (%d)" % (table_name, start))

    def reset_sequence(self, name, start=1):
        """Resets a sequence.
        """
        table_name = self.prefix + name
        self.execute("UPDATE %s SET last_value=0" % table_name)

    def increment(self, name):
        """Increments a sequence.
        """
        table_name = self.prefix + name
        self.execute(
            "UPDATE %s SET last_value=LAST_INSERT_ID(last_value+1)" %
            table_name)
        rows = self.execute("SELECT LAST_INSERT_ID()", fetch=1)
        return rows[0][0]
