##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Zope 2 object classification.

$Id: classifier.py,v 1.1 2004/06/28 18:08:30 johanc Exp $
"""

from mimetypes import guess_extension

from Acquisition import aq_base
from OFS.ObjectManager import ObjectManager

from apelib.core.interfaces import IConfigurableClassifier
from apelib.core.interfaces import SerializationError, DeserializationError

# guess_extension() is useful, but it's unoptimized and sometimes
# chooses strange extensions.  fixed_extensions does nothing other than
# suggest a filename extension given a content type.  It contains
# some better defaults.
fixed_extensions = {
    'text/plain': '.txt',
    'text/html':  '.html',
    'image/png':  '.png',
    'image/jpg':  '.jpg',
    'image/jpeg': '.jpg',
    'image/gif':  '.gif',
    'application/octet-stream': '',  # No extension--too overloaded.
    }

generic_classifications = (
    'directory', 'file', 'folder_object', 'file_object', 'basepath')


class Classifier:
    """A classifier with some minimal Zope 2 extensions.
    """
    __implements__ = IConfigurableClassifier
    gateway = None

    def __init__(self, gw=None):
        self.gateway = gw
        self.oid_to_mapper = {}
        self.class_to_mapper = {}
        self.ext_to_mapper = {}
        self.generic_to_mapper = {}
        self.options = {}  # { (mapper_name, option) -> value }

    def register(self, condition, value, mapper_name):
        value = str(value)  # Avoid unicode
        if condition == 'extension':
            self.ext_to_mapper[value] = mapper_name
        elif condition == 'class':
            self.class_to_mapper[value] = mapper_name
        elif condition == 'generic':
            assert value in generic_classifications, value
            self.generic_to_mapper[value] = mapper_name
        elif condition == 'oid':
            self.oid_to_mapper[value] = mapper_name
        else:
            raise ValueError('Unknown classification condition: %s'
                             % repr(condition))

    def set_option(self, mapper_name, option, value):
        assert option in ('default_extension', 'content_type_attr'), option
        self.options[(mapper_name, option)] = value


    def classify_object(self, event):
        """Chooses a mapper and classification for storing an object.
        """
        mapper_name = self.oid_to_mapper.get(event.oid)
        if mapper_name is not None:
            return {'mapper_name': mapper_name}
        klass = event.obj.__class__
        class_name = '%s.%s' % (klass.__module__, klass.__name__)
        classification = {'class_name': class_name}
        mapper_name = self.class_to_mapper.get(class_name)
        if mapper_name is None:
            folderish = isinstance(event.obj, ObjectManager)
            # Store in a generic format
            if folderish:
                generic = 'folder_object'
            else:
                generic = 'file_object'
            mapper_name = self.generic_to_mapper.get(generic)
        if mapper_name is None:
            raise ClassificationError(
                'No mapper known for class %s' % repr(class_name))
        cta = self.options.get((mapper_name, 'content_type_attr'))
        if cta is not None:
            ct = str(getattr(aq_base(event.obj), cta, None))
            ext = fixed_extensions.get(ct)
            if ext is None:
                ext = guess_extension(ct)
        else:
            ext = self.options.get((mapper_name, 'default_extension'))
        if ext:
            classification['extension'] = ext

        classification['mapper_name'] = mapper_name
        return classification


    def classify_state(self, event):
        """Chooses a mapper and classification for loading an object.
        """
        mapper_name = self.oid_to_mapper.get(event.oid)
        if mapper_name is not None:
            return {'mapper_name': mapper_name}
        classification, serial = self.gateway.load(event)
        class_name = classification.get('class_name')
        if class_name and ':' in class_name:
            # Backward compatibility
            class_name = class_name.replace(':', '.')
            classification['class_name'] = class_name
        mapper_name = classification.get('mapper_name')
        if mapper_name is None:
            # The choice of mapper is not stored explicitly.  Choose
            # one based on several criteria.
            if 1:
                # bw compat: look for certain meta_types.
                mt = classification.get('meta_type')
                if mt == '(folderish object)':
                    mapper_name = self.generic_to_mapper.get('folder_object')
                elif mt == '(fileish object)':
                    mapper_name = self.generic_to_mapper.get('file_object')
            if mapper_name is None:
                if class_name is not None:
                    mapper_name = self.class_to_mapper.get(class_name)
            if mapper_name is None:
                subpath = classification.get('subpath')
                if subpath is not None and not subpath:
                    # Application base
                    mapper_name = self.generic_to_mapper.get('basepath')
            if mapper_name is None:
                t = classification.get('node_type')
                if t == 'd':
                    # Directory
                    mapper_name = self.generic_to_mapper.get('directory')
                elif t == 'f':
                    # File
                    ext = classification.get('extension')
                    if ext:
                        if not ext.startswith('.'):
                            ext = '.' + ext
                        mapper_name = self.ext_to_mapper.get(ext.lower())
                    if not mapper_name:
                        mapper_name = self.generic_to_mapper.get('file')
        if mapper_name is None:
            raise ClassificationError(
                'No mapper known for oid %s' % repr(event.oid))

        classification['mapper_name'] = mapper_name
        return classification
